package de.uniulm.mathematik.typo.fonts.afm;

import java.util.*;

/**
 * Instances of this class represent tables of metric informations
 * for glyphs where, as far as defined, all encodings and all names
 * are unique.
 */
public class CharacterMetricsTable {
   private HashMap<Integer, CharacterMetrics> byCode;
   private HashMap<String, CharacterMetrics> byName;
   private int min, max;
   private int count;

   /**
    * Inner class for exceptions raised by the CharacterMetricsTable class.
    */
   public static class CharacterMetricsTableException extends Exception {
      /**
       * The key.
       */
      public final String key;

      /**
       * A printable error message.
       */
      public final String message;

      /**
       * Construct an exception object encapsulating a
       * key and a printable message within a throw statement.
       * @param key encoding value or glyph name.
       * @param message printable error message.
       */
      public CharacterMetricsTableException(String key, String message) {
	 this.key = key;
	 this.message = message;
      }

      /**
       * Return a string containing an error message which
       * includes both, the token and the error message.
       * @return the string with the error message.
       */
      public String toString() {
	 if (key != null) {
	    return message + ": " + key;
	 } else {
	    return message;
	 }
      }
   } // class CharacterMetricsTableException

   /**
    * Construct a character metrics table which is initially empty.
    */
   public CharacterMetricsTable() {
      byCode = new HashMap<Integer, CharacterMetrics>();
      byName = new HashMap<String, CharacterMetrics>();
      min = -1; max = -1; count = 0;
   }

   /**
    * Add an individual character metrics to the table which must
    * not be in conflict with already existing entries.
    * @param cm individual character metrics to be added.
    */
   public void add(CharacterMetrics cm)
	 throws CharacterMetricsTableException {
      String name = cm.getName();
      if (cm.getCode() < 0 && name == null) {
	 throw new CharacterMetricsTableException(cm.toString(),
	    "neither code value nor name defined");
      }
      if (cm.getCode() >= 0) {
	 Integer code = new Integer(cm.getCode());
	 if (byCode.containsKey(code)) {
	    throw new CharacterMetricsTableException(code.toString(),
	       "multiply defined code");
	 }
	 byCode.put(code, cm);
	 if (code > max) {
	    max = code;
	 }
	 if (code < min || min == -1) {
	    min = code;
	 }
      }
      if (name != null) {
	 if (byName.containsKey(name)) {
	    throw new CharacterMetricsTableException(name,
	       "multiply defined name");
	 }
	 byName.put(name, cm);
      }
      ++count;
   }

   /**
    * Returns true, if this table includes an entry for the given
    * encoding.
    * @param code encoding value, in the range 0..255.
    */
   public boolean defined(int code) {
      return byCode.containsKey(new Integer(code));
   }

   /**
    * Returns true, if this table includes an entry with the given name.
    * @param name name of the glyph.
    */
   public boolean defined(String name) {
      return byName.containsKey(name);
   }

   /**
    * Return the character metrics entry for the given encoding.
    * Null is returned, if there is no glyph with the given encoding value.
    * @param code encoding value, in the range 0..255.
    */
   public CharacterMetrics get(int code) {
      return byCode.get(new Integer(code));
   }

   /**
    * Return the character metrics entry for the glyph with the given
    * name. Null is returned, if there is no glyph with the given name.
    * @param name name of the glyph.
    */
   public CharacterMetrics get(String name) {
      return byName.get(name);
   }

   /**
    * Return the lowest encoding value encountered so far.
    * -1 is returned, if no encodings have been seen yet.
    */
   public int minCode() {
      return min;
   }

   /**
    * Return the highest encoding value encounted so far.
    * -1 is returned, if no encodings have been seen yet.
    */
   public int maxCode() {
      return max;
   }

   /**
    * Return the number of entries within the table.
    */
   public int length() {
      return count;
   }

   /**
    * Return a representation of the internal data structures
    * as printable string.
    */
   public String toString() {
      return
	 "character metrics by code value: " + byCode.toString() + "\n" +
	 "character metrics by name: " + byName.toString() + "\n" +
	 "minimal code value: " + min + "\n" +
	 "maximal code value: " + max + "\n" +
	 "length: " + count + "\n";
   }
} // CharacterMetricsTable
