package de.uniulm.mathematik.typo.fonts.afm;

import java.io.*;
import java.lang.Integer;

/**
 * Lexical analysis for Adobe Font Metric files.
 * See Adobe Font Metrics File Specification, chapter 3.
 * <P>
 * Under the assumption that the source file is conformant, a
 * sequence of lexical tokens generated by an instance of this
 * class follows following grammar:
 * <P>
 * <UL>
 * <LI>FontMetrics --&gt; START VALUE ListOfProperties END</LI>
 * <LI>ListOfProperties --&gt;</LI>
 * <LI>ListOfProperties --&gt; Property ListOfProperties</LI>
 * <LI>Property --&gt; SimpleProperty</LI>
 * <LI>Property --&gt; Block</LI>
 * <LI>Property --&gt; KernData</LI>
 * <LI>SimpleProperty --&gt; PROPERTY OptionalValue</LI>
 * <LI>KernData --&gt; KERN_DATA_START ListOfBlocks KERN_DATA_END</LI>
 * <LI>ListOfBlocks --&gt; Block</LI>
 * <LI>ListOfBlocks --&gt; Block ListOfBlocks</LI>
 * <LI>Block --&gt; BLOCK_START OptionalValue ListOfBlockLines BLOCK_END</LI>
 * <LI>OptionalValue --&gt;</LI>
 * <LI>OptionalValue --&gt; VALUE</LI>
 * <LI>ListOfBlockLines --&gt;</LI>
 * <LI>ListOfBlockLines --&gt; BLOCK_LINE ListOfBlockLines</LI>
 * </UL>
 */

public class Lex {
   private final BufferedReader in; // input stream
   private boolean eof; // eof detected?
   private boolean error; // error detected?
   private char ch; // current character read
   private final static int EXPECT_PROPERTY = 0;
   private final static int EXPECT_VALUE = 1;
   private final static int EXPECT_VALUE_THEN_BLOCK = 2;
   private final static int EXPECT_BLOCK = 3;
   private int state;
   private String blockname;
   private String block;
   private int blocklength;

   /**
    * Inner class whose instances represent individual tokens. A
    * token is a tuple consisting of a symbol and an associated
    * object.
    */
   public static class Token {
      /**
       * Symbol representing an errorneous state. This state is sticky,
       * i.e. once ERROR has been returned as symbol, only ERROR will
       * be returned subsequently.
       */
      public final static int ERROR = 0;
      /**
       * Symbol representing the end of the input source. The associated
       * state is sticky, i.e. once EOF has been returned as symbol, only
       * EOF will be returned subsequently.
       */
      public final static int EOF = 1;
      /**
       * Symbol representing the keyword StartFontMetrics.
       */
      public final static int START = 2;
      /**
       * Symbol representing the keyword EndFontMetrics.
       */
      public final static int END = 3;
      /**
       * Keyword symbol that represents the name of a property. The
       * associated object is a string with the name of the property.
       */
      public final static int PROPERTY = 4;
      /**
       * Comment symbol which is returned by getTokenOrComment but
       * not by getToken. The associated object is a string with the
       * text of the comment.
       */
      public final static int COMMENT = 5;
      /**
       * Property values and some of the blocks have a value. The
       * associated object is a string that contains the text following
       * the keyword up to but not including the line terminator.
       */
      public final static int VALUE = 6; // comes with the value of the key
      /**
       * Symbol representing a StartXXX keyword which is to be balanced
       * by an EndXXX keyword. Intermediate lines remain unparsed and
       * are represented as BLOCK_LINE tokens. The end of the block is
       * signalled by a BLOCK_END symbol. The associated object contains
       * the name of the block without the leading "Start", e.g. "CharMetrics".
       * The BLOCK_START symbol is possibly followed by a VALUE token.
       */
      public final static int BLOCK_START = 7;
      /**
       * Symbol representing an individual line of input within a block.
       * The associated object provides the unparsed line of source text.
       */
      public final static int BLOCK_LINE = 8;
      /**
       * Symbol signalling the end of a block.
       */
      public final static int BLOCK_END = 9;
      /**
       * Symbol representing the StartKernData keyword.
       */
      public final static int KERN_DATA_START = 10;
      /**
       * Symbol representing the EndKernData keyword.
       */
      public final static int KERN_DATA_END = 11;
      /**
       * Table of strings for the individual symbols.
       */
      public final static String SYM_NAME[] = {
	 "ERROR", "EOF", "START", "END", "PROPERTY", "COMMENT", "VALUE",
	 "BLOCK_START", "BLOCK_LINE", "BLOCK_END",
	 "KERN_DATA_START", "KERN_DATA_END"
      };

      /**
       * The symbol of this token. This is a value between ERROR
       * and KERN_DATA_END.
       */
      public final int symbol;

      /**
       * Associated object, if any.
       */
      public final Object object;

      /**
       * Construct a constant token object.
       * @param symbol its symbol, must be between ERROR and KERN_DATA_END.
       * @param object its associated object (may be null).
       */
      public Token(int symbol, Object object) {
	 this.symbol = symbol;
	 this.object = object;
      }

      /**
       * Check if it is a valid symbol, i.e. not EOF and not ERROR.
       * @return true, if it is neither EOF nor ERROR.
       */
      public boolean valid() {
	 return symbol > EOF;
      }

      /**
       * Generate a string representation that might be useful
       * for error messages. It includes the name of the symbol
       * and, if non-null, the object.
       * @return a string, ready to be printed.
       */
      public String toString() {
	 if (object != null) {
	    return SYM_NAME[symbol] + " \"" + object + "\"";
	 } else {
	    return SYM_NAME[symbol];
	 }
      }
   } // class Token

   private Token nextToken;

   /**
    * Create a lex object that scans lexical tokens for
    * Adobe Font Metrics (AFM).
    *
    * @param in line buffered input stream.
    */
   public Lex(BufferedReader in) {
      this.in = in;

      /*
       * Initialize the engine. As we are always looking
       * one character ahead we have to pick up the next
       * one here using next().
       */
      eof = false;
      error = false;
      state = EXPECT_PROPERTY;
      nextToken = null;
      next();
   }

   /**
    * Check if this lex object continues to read from its input.
    * @return true unless end of input or an error state is reached.
    *         Once false is returned, it remains false for all
    *         further invocations.
    */
   public boolean valid() {
      return !eof && !error;
   }

   private Token getSimpleToken() {
      while (!error && !eof && Character.isWhitespace(ch)) {
	 if (state == EXPECT_VALUE && ch == '\n') {
	    state = EXPECT_PROPERTY;
	 }
	 next();
      }
      if (error) {
	 return new Token(Token.ERROR, null);
      }
      if (eof) {
	 return new Token(Token.EOF, null);
      }

      switch(state) {
	 case EXPECT_PROPERTY:
	    if (!Character.isLetter(ch)) {
	       error = true; return new Token(Token.ERROR, null);
	    }
	    String keyname = (new Character(ch)).toString();
	    next();
	    while (!eof && !error && (!Character.isWhitespace(ch))) {
	       keyname += ch;
	       next();
	    }
	    state = EXPECT_VALUE;
	    if (keyname.equals("StartFontMetrics")) {
	       return new Token(Token.START, keyname);
	    } else if (keyname.equals("EndFontMetrics")) {
	       return new Token(Token.END, keyname);
	    } else if (keyname.equals("StartKernData")) {
	       return new Token(Token.KERN_DATA_START, keyname);
	    } else if (keyname.equals("EndKernData")) {
	       return new Token(Token.KERN_DATA_END, keyname);
	    } else if (keyname.length() > 5 &&
		  keyname.substring(0,5).equals("Start")) {
	       state = EXPECT_VALUE_THEN_BLOCK;
	       keyname = keyname.substring(5);
	       if (keyname.equals("KernPairs0") ||
		     keyname.equals("KernPairs1")) {
		  // StartKernPairs0 and StartKernPairs1 are
		  // unfortunately balanced by EndKernPairs
		  blockname = "KernPairs";
	       } else {
		  blockname = keyname;
	       }
	       return new Token(Token.BLOCK_START, keyname);
	    } else {
	       return new Token(Token.PROPERTY, keyname);
	    }
	 case EXPECT_VALUE:
	 case EXPECT_VALUE_THEN_BLOCK:
	    String param = (new Character(ch)).toString();
	    next();
	    while (!eof && !error && ch != '\n') {
	       param += ch;
	       next();
	    }
	    if (state == EXPECT_VALUE_THEN_BLOCK) {
	       state = EXPECT_BLOCK;
	       try {
		  blocklength = java.lang.Integer.parseInt(param, 10);
	       } catch (NumberFormatException e) {
		  error = true;
		  return new Token(Token.ERROR, null);
	       }
	    } else {
	       state = EXPECT_PROPERTY;
	    }
	    return new Token(Token.VALUE, param);
	 case EXPECT_BLOCK:
	    String line = "";
	    while (!eof && !error && ch != '\n') {
	       line += ch;
	       next();
	    }
	    if (eof || error) {
	       return new Token(Token.ERROR, null);
	    }
	    if (line.equals("End" + blockname)) {
	       if (blocklength > 0) {
		  error = true;
		  return new Token(Token.ERROR, null);
	       }
	       state = EXPECT_PROPERTY;
	       return new Token(Token.BLOCK_END, blockname);
	    }
	    if (blocklength < 0) {
	       error = true;
	       return new Token(Token.ERROR, null);
	    }
	    --blocklength;
	    return new Token(Token.BLOCK_LINE, line);
	 default:
	    error = true;
	    return new Token(Token.ERROR, null);
      }
   }

   private Token getNextToken() {
      Token token;
      if (nextToken != null) {
	 token = nextToken; nextToken = null;
      } else {
	 token = getSimpleToken();
      }
      return token;
   }

   /**
    * Return the next token from the input source. Comment tokens
    * are not suppressed.
    */
   public Token getTokenOrComment() {
      Token token = getNextToken();
      if (token.symbol == Token.PROPERTY && token.object.equals("Comment")) {
	 Token comment = getNextToken();
	 if (comment.symbol != Token.VALUE) {
	    nextToken = comment;
	    return new Token(Token.COMMENT, "");
	 }
	 return new Token(Token.COMMENT, comment.object);
      }
      return token;
   }

   /**
    * Read and return the next token from the input excluding comments.
    * Note however that no further read operations are performed
    * once a error or end of input state is reached.
    * @return next input token.
    */
   public Token getToken() {
      Token token = getTokenOrComment();
      while (token.symbol == Token.COMMENT) {
	 token = getTokenOrComment();
      }
      return token;
   }

   /**
    * Read the next character from the input and
    * assign it to ch.
    */
   private void next() {
      if (!eof) {
	 int nextch;
	 try {
	    nextch = in.read();
	 } catch (Exception e) {
	    error = true;
	    nextch = -1;
	 }
	 if (nextch == -1) {
	    eof = true;
	 } else {
	    ch = (char) nextch;
	 }
      }
   }
} // class Lex
