package de.uniulm.mathematik.typo.fonts.afm;

import java.util.*;

/**
 * An instance of a property table maintains any number of
 * property/value pairs whose property keys are unique.
 * Properties may be added any time but they can be neither
 * removed nor updated.
 */
public class PropertyTable {
   private HashMap<String, Object> table;

   /**
    * Inner class representing exceptions generated by
    * the PropertyTable class.
    */
   public static class PropertyTableException extends Exception {
      /**
       * The property/value pair.
       */
      public final String property;
      public final Object value;

      /**
       * A printable error message.
       */
      public final String message;

      /**
       * Construct an exception object encapsulating a
       * token and a printable message within a throw statement.
       * @param property name of property key.
       * @param value associated value.
       * @param message printable error message.
       */
      public PropertyTableException(String property, Object value,
                                    String message) {
	 this.property = property;
	 this.value = value;
	 this.message = message;
      }

      /**
       * Return a string containing an error message which
       * includes both, the message and the property/value pair.
       * @return the string with the error message.
       */
      public String toString() {
	 if (value != null) {
	    return message + ": " + property + " " + value;
	 } else {
	    return message + ": " + property;
	 }
      }
   } // class PropertyTableException

   /**
    * Construct a property table that is initially empty.
    */
   public PropertyTable() {
      table = new HashMap<String, Object>();
   }

   /**
    * Add a property/value pair to the table. The property must
    * not be in conflict to previous entries of the table.
    * @param property name of property.
    * @param value associated value, may be null. 
    */
   public void add(String property, Object value)
	 throws PropertyTableException {
      if (table.containsKey(property)) {
	 throw new PropertyTableException(property, value,
	    "property multiply defined");
      }
      table.put(property, value);
   }

   /**
    * Returns true, if a property with the given name is
    * defined in the table.
    */
   public boolean defined(String property) {
      return table.containsKey(property);
   }

   /**
    * Return the value associated with property. The given
    * property must exist. The returned can may be null.
    * @param property name of property.
    */
   public Object get(String property) throws PropertyTableException {
      if (!table.containsKey(property)) {
	 throw new PropertyTableException(property, null, "undefined");
      }
      return table.get(property);
   }

   /**
    * Return internal data structure as printable string.
    */
   public String toString() {
      return table.toString();
   }
} // PropertyTable
