package de.uniulm.mathematik.typo.hyphen;

import java.util.*;
import de.uniulm.mathematik.tries.*;

/**
 * An instance of this class allows to parse hyphenation patterns
 * and to create a trie data structure for it.
 * <P/>
 * The syntax of hyphenation patterns follows that of TeX:
 * <UL>
 * <LI>A pattern is a sequence of letters, digits, possibly
 *     with a dot at the beginning and/or the end of the sequence.</LI>
 * <LI>Dots match the beginning or the end of a to be hyphenated word.</LI>
 * <LI>Letters are taken literally. All Unicode codepoints are accepted
 *     but Character.isLetter() shall return true for them.</LI>
 * <LI>Digits represent possible hyphenation positions. The position
 *     is taken relatively to the beginning of the pattern and the
 *     value of the digit specifies how strongly a hyphenation at
 *     that point is to be desired (odd value) or to be suppressed
 *     (even value).
 * </UL>
 * <P/>
 * The resulting trie uses the sequence of letters including
 * the optional dots as key and an object of type HyphenationEntry
 * as value. A hyphenation entry is a sequence of hyphenation points
 * which are tuples consisting of a position and a value.
 * <P/>
 * This implementation compresses the data structures for
 * hyphenation entries used for a trie by eliminating common suffixes.
 */
public class HyphenationTableConstructor {
   private HashMap<HyphenationEntry, HyphenationEntry> entries;
   private HashedTrie<HyphenationEntry> trie;

   /**
    * Create an instance of this class that starts with an
    * empty trie.
    */
   public HyphenationTableConstructor() {
      entries = new HashMap<HyphenationEntry, HyphenationEntry>();
      trie = new HashedTrie<HyphenationEntry>();
   }

   /**
    * Add a rule to the trie which must conform to the syntax
    * of hyphenation patterns (see above).
    */
   public void add(String rule) {
      // extract codepoints out of rule and ...
      //   - move them into codepoints[] and
      //   - add non-digits to word
      int len = rule.codePointCount(0, rule.length());
      int[] codepoints = new int[len];
      int index = 0;
      StringBuffer word = new StringBuffer();
      for (int i = 0; i < rule.length(); i = rule.offsetByCodePoints(i, 1)) {
	 int ch = rule.codePointAt(i);
	 codepoints[index++] = ch;
	 if (!Character.isDigit(ch)) {
	    word.appendCodePoint(ch);
	 }
      }

      // create associated hyphenation entry
      HyphenationEntry entry = null;
      int pos = word.length();
      for (int i = len-1; i >= 0; --i) {
	 int ch = codepoints[i];
	 if (Character.isDigit(ch)) {
	    int value = Character.getNumericValue(ch);
	    entry = new HyphenationEntry(pos, value, entry);
	    HyphenationEntry other = entries.get(entry);
	    if (other != null) {
	       entry = other;
	    }
	 } else {
	    --pos;
	 }
      }
      assert entry != null;
      trie.insert(word.toString(), entry);
   }

   /**
    * Finish the trie construction and return an trie object
    * that supports the TrieReader interface.
    * @see TrieReader
    */
   public TrieReader<HyphenationEntry> finish() {
      return trie;
   }
}
