package de.uniulm.mathematik.typo.items;

import java.util.*;

/**
 * This class allows to join several ItemWrapper objects into
 * one wrapper object.
 */
public class CombinedWrapper implements ItemWrapper {
   private LinkedList<ItemWrapper> wrappers;

   /**
    * Construct a combined wrapper with an initially empty list
    * of wrappers. More wrappers can be added through the add()
    * method.
    */
   public CombinedWrapper() {
      wrappers = new LinkedList<ItemWrapper>();
   }

   /**
    * Construct a combined wrapper with just one wrapper object.
    * More wrappers can be added through the add() method.
    */
   public CombinedWrapper(ItemWrapper first) {
      this(); add(first);
   }

   /**
    * Construct a combined wrapper with the given two wrappers.
    * @param first to be wrapped items will be processed by this
    *              wrapper first.
    * @param second this wrapper comes second after the first wrapper.
    */
   public CombinedWrapper(ItemWrapper first, ItemWrapper second) {
      this(); add(first); add(second);
   }

   /**
    * Construct a combined wrapper with the given three wrappers.
    */
   public CombinedWrapper(ItemWrapper first, ItemWrapper second,
	 ItemWrapper third) {
      this(); add(first); add(second); add(third);
   }

   /**
    * Add an additional wrapper to the combined wrapper. The
    * individual wrappers are applied in the order of their
    * appearance.
    */
   public void add(ItemWrapper wrapper) {
      assert(wrapper != null);
      wrappers.addLast(wrapper);
   }

   /**
    * Apply all wrappers on item and return the resulting wrapper item.
    */
   public Item wrap(Item item) {
      Item result = item;
      for (ItemWrapper wrapper: wrappers) {
	 result = wrapper.wrap(result);
      }
      return result;
   }
} // class CombinedWrapper
