package de.uniulm.mathematik.typo.items;

import java.util.*;
import de.uniulm.mathematik.typo.fonts.*;

/**
 * A framed box encapsulates another box and frames it.
 */
public class FramedItem extends DelegatingItem {
   private Width origwidth;

   /**
    * Construct a wrapped item that represents the original item
    * with a visable representation of the box.
    */
   public FramedItem(Item item) {
      super(item);
      origwidth = new Width(item);
   }

   /**
    * Draw the wrapped item along with
    * <UL>
    * <LI>a visible box in red color in case of regular boxes,
    * <LI>a cyan box in case of kerning boxes,
    * <LI>a vertical green line in case of penalties, and
    * <LI>a blue box in case of glue items.
    * </UL>
    * <P>
    * If the original item has no height or depth (as in case of all
    * non-regular boxes), the vertical length is derived from the
    * metrics of the actually selected font of the context object. If
    * there is no current font defined yet, a height of 12 and a depth
    * of 2 are assumed.
    * <P>
    * In case of glue items, the original width is remembered. The
    * blue box is then drawn with the original width but with a base
    * line that extends to the actual width.
    */
   public StringBuffer genPostScript(PostScriptContext context) {
      StringBuffer result = new StringBuffer("");
      result.append(context.gsave());
      result.append("currentpoint newpath moveto\n");
      if (item.isBox() && item.getDepth() + item.getHeight() > 0) {
	 result.append("1 0 0 setrgbcolor\n");
	 if (item.getDepth() != 0) {
	    result.append("0 " + ((double) -item.getDepth() / 1000) +
	       " rmoveto\n");
	 }
	 result.append(((double) item.getWidth() / 1000) + " 0 rlineto\n");
	 result.append("0 " +
	    ((double) (item.getDepth() + item.getHeight()) / 1000) +
	    " rlineto\n");
	 result.append(((double) -item.getWidth() / 1000) +
	    " 0 rlineto\n");
	 result.append("closepath\n");
      } else {
	 FontMetrics fm = context.getCurrentFont();
	 double height; double depth;
	 if (fm == null) {
	    height = 12; depth = 2;
	 } else {
	    int size = context.getCurrentSize();
	    height = fm.getAscender() * size; height += height / 5;
	    depth = - fm.getDescender() * size; depth += depth / 2;
	    height /= 1000; depth /= 1000;
	 }
	 if (depth > 0) {
	    result.append("0 " + (-depth) + " rmoveto\n");
	 }
	 if (item.isPenalty()) {
	    result.append("0 1 0 setrgbcolor\n");
	    result.append("0 " + (depth + height) + " rlineto\n");
	 } else {
	    assert(item.isGlue() || item.isBox());
	    if (item.isBox()) {
	       result.append("0 1 1 setrgbcolor\n");
	    } else {
	       result.append("0 0 1 setrgbcolor\n");
	    }
	    int delta = item.getWidth() - origwidth.getWidth();
	    if (delta != 0) {
	       result.append(((double) item.getWidth() / 1000) +
		  " 0 rlineto\n");
	       // ... and center the following box
	       result.append(((double) delta / 2000 -
		  (double) item.getWidth() / 1000) + " 0 rmoveto\n");
	    }
	    result.append(((double) origwidth.getWidth() / 1000) +
	       " 0 rlineto\n");
	    result.append("0 " + (depth + height) + " rlineto\n");
	    result.append(((double) - origwidth.getWidth() / 1000) +
	       " 0 rlineto\n");
	    result.append("closepath\n");
	 }
      }
      result.append("stroke\n");
      result.append(context.grestore());
      result.append(item.genPostScript(context));
      return result;
   }
} // class FramedItem
