package de.uniulm.mathematik.typo.items;

import de.uniulm.mathematik.typo.items.*;
import java.util.*;

/**
 * A horizontal box is a closed sequence of individual
 * items that share the same baseline. Horizontal boxes
 * cannot be split.
 */
public class HorizontalBox extends Box {
   private LinkedList<Item> items;
   private int width;
   private int height;
   private int depth;

   /**
    * Construct an initially empty horizontal box.
    */
   public HorizontalBox() {
      items = new LinkedList<Item>();
      width = 0;
      height = 0;
      depth = 0;
   }

   /**
    * Construct a horizontal box from the given sequence.
    */
   public HorizontalBox(HorizontalSequence hseq) {
      this();
      for (Item item: hseq) {
	 add(item);
      }
   }

   /**
    * Add the given item to the sequence of items such that
    * its reference point (beginning point of the base line)
    * will be positioned at the end point of the base line
    * of the last item added to this horizontal box.
    * @param item adds this item to the horizontal box.
    */
   public void add(Item item) {
      items.addLast(item);
      width += item.getWidth();
      if (item.getHeight() > height) {
	 height = item.getHeight();
      }
      if (item.getDepth() > depth) {
	 depth = item.getDepth();
      }
   }

   /**
    * Returns the width of the horizontal box in units (1/1000 points)
    * which equals the sum of all individual widths of the items added so far.
    */
   public int getWidth() {
      return width;
   }

   /**
    * Returns the height of the horizontal box in units (1/1000 points)
    * which equals the maximal height of all items added so far.
    */
   public int getHeight() {
      return height;
   }

   /**
    * Returns the depth of the horizontal box in units (1/1000 points)
    * which equals the maximal depth of all items added so far.
    */
   public int getDepth() {
      return depth;
   }

   /**
    * Invoke the genPostScript() method on all items belonging
    * to this horizontal box, beginning with the first box and
    * finishing with the last box added. The concatenated result
    * of all outputs is returned.
    */
   public StringBuffer genPostScript(PostScriptContext context) {
      StringBuffer result = new StringBuffer("");
      for (Item item:items) {
	 result.append(item.genPostScript(context));
      }
      return result;
   }

   public String toString() {
      StringBuffer result = new StringBuffer("");
      result.append("horizontal box of width = " + width + ":\n");
      for (Item item: items) {
	 result.append(item.toString() + "\n");
      }
      result.append("end of horizontal box\n");
      return result.toString();
   }
} // class HorizontalBox
