package de.uniulm.mathematik.typo.items;

import java.lang.*;

/**
 * Horizontal sequences are horizontally ordered lists of items
 * that are to be stuffed into a horizontal box or to be broken
 * into lines. Individual items that are added to a sequence
 * are either considered as breakpoints or regular items. Following
 * Knuth's definition, legal breakpoints are
 * <UL>
 * <LI>penalties with a penalty value less than INFINITY and
 * <LI>glue boxes that follow a regular box.
 * </UL>
 * <P>
 * A word is a sequence of items that are no breakpoints that is
 * followed by a breakpoint but does not include it. Using
 * getWords(), it is possible to iterate through all words. Note,
 * however, that the last sequence of non-breakpoints is not
 * returned if the final breakpoint is missing.
 */
public interface HorizontalSequence extends Iterable<Item> {

   /**
    * Append the given item at the end of the horizontal sequence.
    * Please note that kerning boxes at the beginning of a sequence
    * are ignored.
    * @param item to be added item.
    */
   public void add(Item item);

   /**
    * Append the given items out of the given horizontal sequence to
    * the end.
    * @param hseq to be added sequence.
    */
   public void add(HorizontalSequence hseq);

   /**
    * Return the overall horizontal width of the sequence. This
    * value is not updated when individual items of the sequence
    * are shrank or stretched. 
    */
   public Width getWidth();

   /**
    * Iterate through all individual words which are represented
    * as subsequences, breakpoints are not included. Note that
    * words can be empty in case of a breakpoint that immediately
    * follows another breakpoint. The missing breakpoint can be
    * retrieved using the getFollowingBreakpoint() method.
    */
   public IterableIterator<HorizontalSequence> getWords();

   /**
    * Return the breakpoint following a subsequence. As getWords()
    * returns no breakpoints this method allows to access the
    * breakpoint following a word. Null is returned for sequences
    * which are not subsequences or subsequences which got extended
    * using one of the add methods.
    */
   public Item getFollowingBreakpoint();

   /**
    * Return glue items, i.e. items that can be shrunk or extended.
    */
   public IterableIterator<Item> getGlueItems();

   /**
    * Return an independent cloned copy of this horizontal sequence.
    * Please note that just the collection is cloned but not the
    * individual items.
    */
   public HorizontalSequence clone();
} // class HorizontalSequence
