package de.uniulm.mathematik.typo.items;

import java.lang.*;
import de.uniulm.mathematik.typo.fonts.*;

/**
 * Instances of this class allow to coordinate the PostScript output
 * generation such that the PostScript code becomes simpler and more
 * compact. Among the supported features is the coordinated string
 * generation for the show operator.
 */
public class PostScriptContext {
   private FontMetrics currentFont;
   private int currentSize;
   private boolean inString;
   private boolean newcontext;

   /**
    * Construct an initial context with no current font.
    */
   public PostScriptContext() {
      currentFont = null;
      currentSize = 0;
      inString = false;
      newcontext = false;
   }

   /**
    * Return the current font, if any.
    */
   public FontMetrics getCurrentFont() {
      return currentFont;
   }
   /**
    * Return the size of the current font, if any.
    */
   public int getCurrentSize() {
      return currentSize;
   }
   /**
    * Returns true, if we are in the middle of a string generation,
    * i.e. if closeString() has to be called before any text not
    * generated by this class is added to the PostScript code.
    */
   public boolean insideString() {
      return inString;
   }

   /**
    * Close the current string and apply the show operator to it,
    * if there is currently an open string.
    */
   public StringBuffer closeString() {
      StringBuffer result = new StringBuffer("");
      if (inString) {
	 result.append(") show\n");
	 inString = false;
      }
      return result;
   }

   /**
    * Emit the invocation of the gsave operator. closeString() is
    * invoked implicitly.
    */
   public StringBuffer gsave() {
      StringBuffer result = new StringBuffer("");
      result.append(closeString());
      result.append("gsave\n");
      return result;
   }

   /**
    * Emit the invocation of the grestore operator. closeString() is
    * invoked implicitly.
    */
   public StringBuffer grestore() {
      StringBuffer result = new StringBuffer("");
      result.append(closeString());
      result.append("grestore\n");
      newcontext = true;
      return result;
   }

   /**
    * Switch to the given font.
    */
   public StringBuffer switchTo(FontMetrics fm, int size) {
      StringBuffer result = new StringBuffer("");
      if (newcontext || fm != currentFont || size != currentSize) {
	 result.append(closeString());
	 result.append("/");
	 result.append(fm.getName());
	 result.append(" findfont ");
	 result.append(size);
	 result.append(" scalefont setfont\n");
	 currentFont = fm; currentSize = size;
	 newcontext = false;
      }
      return result;
   }

   /**
    * Add a glyph to be output in the given font in the given size.
    */
   public StringBuffer addTo(FontMetrics fm, int size, int code) {
      StringBuffer result = new StringBuffer("");
      result.append(switchTo(fm, size));
      if (!inString) {
	 inString = true;
	 result.append("(");
      }
      if (code == '(' || code == ')' || code == '\\') {
	 // convert special characters to an octal representation
	 result.append('\\');
	 result.append(Integer.toOctalString(code));
      } else {
	 result.append((char) code);
      }
      return result;
   }
} // class PostScriptContext
