package de.uniulm.mathematik.typo.items;

import de.uniulm.mathematik.typo.fonts.*;

/**
 * A sequencer generates a horizontal sequence out of a sequence
 * of input characters. Following mappings apply:
 * <P>
 * <UL>
 * <LI>Sequences of white space are converted into glue items.
 * <LI>Sequences of dashes are extended by an unflagged penalty item
 *     with a penalty value of 0.
 * <LI>The sequence `\-' is converted into a hyphenation item.
 * <LI>Encoded characters are mapped into glyph boxes. Kerning boxes
 *     are inserted between glyph boxes, whenenver a non-zero kerning
 *     is specified.
 * <LI>All other unencoded characters are ignored.
 * </UL>
 */
public class Sequencer {
   private HorizontalSequence hseq;
   private FontMetrics fm;
   private int size;
   private boolean lastCharDefined;
   private int lastChar;
   private int spaceWidth;
   private int shrinkability;
   private int stretchability;
   private boolean addGlue;
   private ItemWrapper wrapper;

   /**
    * Construct a sequencer that adds all its generated items
    * to the given horizontal sequence. The initial font is
    * undefined.
    * @param hseq the horizontal sequence generated items are added to.
    */
   public Sequencer(HorizontalSequence hseq) {
      this.hseq = hseq;
      fm = null;
      size = 0;
      shrinkability = 0;
      stretchability = 0;
      lastCharDefined = false;
      addGlue = false;
      wrapper = null;
   }

   /**
    * Construct a sequencer that adds all its generated items
    * to the given horizontal sequence.
    * @param hseq the horizontal sequence generated items are added to.
    * @param fm the font metrics of the initial font.
    * @param size the size of the initial font in points.
    */
   public Sequencer(HorizontalSequence hseq, FontMetrics fm, int size) {
      this(hseq);
      this.setFont(fm, size);
   }

   public void setWrapper(ItemWrapper wrapper) {
      this.wrapper = wrapper;
   }

   private void add(Item item) {
      if (wrapper != null) {
	 item = wrapper.wrap(item);
      }
      hseq.add(item);
   }

   private void processLastChar() {
      if (lastCharDefined) {
	 Item item = new NormalizedGlyphBox(fm, size, lastChar);
	 add(item);
	 if (lastChar == '-') {
	    Item penalty = new Penalty(1, true);
	    add(penalty);
	 }
	 lastCharDefined = false;
      } else if (addGlue) {
	 Glue glue = new Glue(spaceWidth * size,
	    stretchability * size, shrinkability * size);
	 add(glue);
	 addGlue = false;
      }
   }

   /**
    * Switch to the given font with the given size.
    * @param fm the font metrics of the initial font.
    * @param size the size of the initial font in points.
    */
   public void setFont(FontMetrics fm, int size) {
      processLastChar();
      assert(fm != null);
      this.fm = fm;
      this.size = size;
      spaceWidth = fm.getWidth(' ');
      shrinkability = spaceWidth / 3;
      stretchability = spaceWidth / 2;
   }

   /**
    * By default, the space width is taken from the metrics of
    * the space character. This method allows to change the
    * metrics of spaces after a font has been selected.
    * @param spaceWidth the width of a space in units (1/1000 points).
    */
   public void setSpaceWidth(int spaceWidth) {
      this.spaceWidth = spaceWidth;
   }

   /**
    * By default, spaces can be shrank by a 1/3 of their regular width.
    * This method allows to override this.
    */
   public void setShrinkability(int shrinkability) {
      this.shrinkability = shrinkability;
   }

   /**
    * By default, spaces can be stretched by a 1/2 of their regular width.
    * This method allows to override this.
    */
   public void setStretchability(int shrinkability) {
      this.stretchability = stretchability;
   }

   /**
    * Add the given character to the horizontal sequence. Kerning boxes
    * are automatically created and inserted, whenever suggested by the
    * font metrics. Sequences of whitespace are converted into glue
    * boxes. This operation is ignored if the encoding vector of the
    * current font does not include the given code.
    */
   public void add(int code) {
      assert(fm != null);
      if (Character.isWhitespace(code)) {
	 if (!addGlue) {
	    processLastChar();
	    addGlue = true;
	 }
      } else if (fm.defined(code)) {
	 boolean suppress = false;
	 if (lastCharDefined) {
	    if (code == '-' && lastChar == '-') {
	       add(new NormalizedGlyphBox(fm, size, lastChar));
	    } else if (lastChar == '\\' && code == '-') {
	       add(new Hyphenation(new NormalizedGlyphBox(fm, size, '-')));
	       lastCharDefined = false; suppress = true;
	    } else {
	       processLastChar();
	       int kerning = fm.getKerning(lastChar, code);
	       if (kerning != 0) {
		  add(new KerningBox(kerning * size));
	       }
	    }
	 } else if (addGlue) {
	    processLastChar();
	 }
	 if (!suppress) {
	    lastChar = code; lastCharDefined = true;
	 }
      }
   }

   /**
    * Finish the horizontal sequence. A final breakpoint is not
    * added to it.
    */
   public void finish() {
      if (lastCharDefined) {
	 processLastChar();
      }
   }

   /**
    * Finish the horizontal sequence and add a glue object
    * to it with infinity stretchability.
    */
   public void finishParagraph() {
      finish();
      add(new Glue(0, Item.INFINITY, 0));
      // add(new Penalty(-Item.INFINITY, false));
   }
} // class Sequencer
