package de.uniulm.mathematik.typo.items;

import java.util.*;

/**
 * Vertical boxes are regular boxes that stack any number of items
 * vertically. The distances between the base lines of the individual
 * lines are constant.
 */

public class VerticalBox extends Box {
   private LinkedList<Item> items;
   private int width;
   private int height;
   private int depth;
   private int baselineskip;

   /**
    * Construct a vertical box.
    * @param baselineskip specifies the vertical distance between
    *                     the baselines of two consecutive members.
    */
   public VerticalBox(int baselineskip) {
      items = new LinkedList<Item>();
      width = 0;
      height = 0;
      depth = 0;
      this.baselineskip = baselineskip;
   }

   /**
    * Add this item to the vertical box, i.e. put it below the
    * last item added with a vertical distance of baselineskip.
    * @param item item to be added.
    */
   public void add(Item item) {
      items.addLast(item);
      if (item.getWidth() > width) {
	 width = item.getWidth();
      }
      height += baselineskip;
      depth = item.getDepth();
   }

   /**
    * Return the maximal width seen so far of all individual
    * items added to this vertical box.
    */
   public int getWidth() {
      return width;
   }

   /**
    * Return baselineskip times the number of items added so far,
    * diminished by the actual depth such that the sum of height
    * and depth are a multiple of baselineskip.
    */
   public int getHeight() {
      return height - depth;
   }

   /**
    * Return the depth of the last item added to the vertical
    * list. Hence, the base line of the entire vertical box
    * equals the baseline of the last item added to that box.
    */
   public int getDepth() {
      return depth;
   }

   /**
    * Emit PostScript text that represent the contents of
    * this vertical box.
    */
   public StringBuffer genPostScript(PostScriptContext context) {
      double bskip = (double) baselineskip / 1000;
      StringBuffer result = new StringBuffer("");
      if (items.size() > 0) {
	 int i = 0;
	 for (Item item:items) {
	    result.append(context.gsave());
	    if (i < items.size()-1) {
	       result.append("0 ");
	       result.append(bskip * (items.size()-1-i));
	       result.append(" rmoveto\n");
	    }
	    result.append(item.genPostScript(context));
	    result.append(context.closeString());
	    result.append(context.grestore());
	    ++i;
	 }
	 result.append((double) width / 1000);
	 result.append(" 0 rmoveto\n");
      }
      return result;
   }
} // class VerticalBox
